# This code is part of the Biopython distribution and governed by its
# license.  Please see the LICENSE file that should have been included
# as part of this package.
#

"""Code to deal with various programs for sequencing and assembly.

This code deals with programs such as Phred, Phrap and Consed -- which provide
utilities for calling bases from sequencing reads, and assembling sequences
into contigs.
"""
